import {
  GridEditable,
  IGridEditable,
  IGridEditableEvents,
} from '@zeedhi/common';
import { IEventParam, IDictionary, IEvent } from '@zeedhi/core';

export interface InterfaceDragDropGridEvent extends IEventParam<DragDropGrid> {
  row: IDictionary;
}

export interface InterfaceDragDropGridEvents<T = InterfaceDragDropGridEvent | IEventParam<IDictionary>> extends IGridEditableEvents<T> {
  onDragMove?: IEvent<T>;
  onDragStart?: IEvent<T>;
  onDragEnd?: IEvent<T>;
}

export interface InterfaceDragDropGrid extends IGridEditable {
  orderColumnName: string;
  events?: InterfaceDragDropGridEvents;
}

export class DragDropGrid extends GridEditable implements InterfaceDragDropGrid {
  /** Name of column to control position using drag drop */
  public orderColumnName: string;

  /** Grid Events */
  public events!: InterfaceDragDropGridEvents;

  /**
   * Grid constructor
   * @param props Grid properties
   */
  constructor(props: InterfaceDragDropGrid) {
    super({
      ...props,
      datasource: {
        ...props.datasource, order: [props.orderColumnName],
      },
    });
    this.orderColumnName = props.orderColumnName;
    this.setDatasourceGetMethod();
  }

  /**
   * Change datasource get method to reflect changes on order
   */
  private setDatasourceGetMethod() {
    const oldDatasourceGet = this.datasource.get.bind(this.datasource);
    this.datasource.get = () => oldDatasourceGet().then((data: IDictionary[]) => {
      data.sort(this.sortData.bind(this));

      return data;
    });
  }

  private sortData(row: IDictionary, nextRow: IDictionary) {
    const rowId = row[this.datasource.uniqueKey];
    const nextRowId = nextRow[this.datasource.uniqueKey];
    const editedRows = (this as IDictionary).editedRows;
    const rowOrder = editedRows[rowId]?.[this.orderColumnName] || row[this.orderColumnName];
    const nextRowOrder = editedRows[nextRowId]?.[this.orderColumnName] || nextRow[this.orderColumnName];
    if (rowOrder < nextRowOrder) return -1;
    if (rowOrder > nextRowOrder) return 1;
    return 0;
  }

  public async cancelEditedRows() {
    return super.cancelEditedRows().then(() => {
      this.datasource.data.sort(this.sortData.bind(this));
    });
  }

  /**
   * Event triggered when dragged element is moving
   * @param event DOM event
   * @param row row being dragged
   * @param element HTML element
   * @returns return false to cancel dragging
   */
  public onDragMove(event: Event, row: IDictionary, element: HTMLElement) {
    if (this.events.onDragMove) {
      return this.events.onDragMove({
        component: this,
        event,
        row,
        element,
      });
    }

    return true;
  }

  /**
   * Event triggered when drag starts
   * @param event DOM event
   * @param row row being dragged
   * @param element HTML element
   */
  public onDragStart(event: Event, row: IDictionary, element: HTMLElement) {
    if (this.events.onDragStart) {
      this.events.onDragStart({
        component: this,
        event,
        row,
        element,
      });
    }
  }

  /**
   * Event triggered when drag ends
   * @param event DOM event
   * @param row row being dragged
   * @param element HTML element
   */
  public onDragEnd(event: Event, row: IDictionary, element: HTMLElement) {
    const { oldIndex, newIndex } = (event as IDictionary);
    const { data, uniqueKey } = this.datasource;
    const editedRows = (this as IDictionary).editedRows;

    let idx = newIndex;
    const rowOldOrder = editedRows[data[idx][uniqueKey]]?.[this.orderColumnName] || row[this.orderColumnName];
    const originalColumnName = `${this.orderColumnName}_original`;
    for (let i = 0; i <= Math.abs(oldIndex - newIndex); i += 1) {
      const nextIdx = idx + (oldIndex > newIndex ? 1 : -1);
      const id = data[idx][uniqueKey];
      const nextId = data[nextIdx]?.[uniqueKey];
      const newRow = {
        ...editedRows[id],
        [originalColumnName]: editedRows[id]?.[originalColumnName] || data[idx][this.orderColumnName],
        [this.orderColumnName]: idx === oldIndex ? rowOldOrder : editedRows[nextId]?.[this.orderColumnName] || data[nextIdx][this.orderColumnName],
        originalRow: data[idx],
      };

      editedRows[data[idx][uniqueKey]] = newRow;

      idx = nextIdx;
    }

    this.editing = true;
    if (this.events.inlineEdit && typeof this.events.inlineEdit === 'function') {
      this.events.inlineEdit({
        event, element, row, column: { name: this.orderColumnName }, component: this,
      });
    }

    if (this.events.onDragEnd) {
      this.events.onDragEnd({
        component: this,
        event,
        row,
        element,
      });
    }
  }
}
